//--------------------------------------------------------------------------------------
// File: iig_fp16_hdr.fx
//
// Desc:  Shaders based on implementation in HDRFormats in DX9 SDK Summer Update 2004
//--------------------------------------------------------------------------------------

// ------------
// Global SAS compliance
// ---------------
int global : SasGlobal
<
	bool SasUiVisible = false;
	int3 SasVersion = {1,0,0};
>;

//--------------------------------------------------------------------------------------
// Global variables
//--------------------------------------------------------------------------------------
float4x4 g_mWorld;					// World matrix for object
float4x4 g_mWorldViewProj;			// World * View * Projection matrix

static const float3 LUMINANCE_VECTOR  = float3(0.2125f, 0.7154f, 0.0721f);
static const float  MODEL_REFLECTIVITY = 0.2f;
static const float  LUM_WHITE = 1.5f;
static const float  BRIGHT_THRESHOLD = 2.5f;

float3 g_vEyePt;
float3 model_color;
float  g_fImageKey;
float  g_fMiddleGray;

float2 g_avSampleOffsets[16];
float4 g_avSampleWeights[16];

textureCUBE g_tCube;

samplerCUBE CubeSampler = 
sampler_state
{
    Texture = <g_tCube>;
    MinFilter = Point;
    MagFilter = Point;
    MipFilter = Point;
};

sampler s0 : register(s0);
sampler s1 : register(s1);
sampler s2 : register(s2);
sampler s3 : register(s3);
sampler s4 : register(s4);
sampler s5 : register(s5);
sampler s6 : register(s6);
sampler s7 : register(s7);

//-----------------------------------------------------------------------------
// Name: SkyboxVS
//-----------------------------------------------------------------------------
struct SkyboxVS_Input
{
    float3 Pos : POSITION;
    float2 Tex : TEXCOORD0;
};

struct SkyboxVS_Output
{
    float4 Pos : POSITION;
    float2 Tex : TEXCOORD0;
};

SkyboxVS_Output SkyboxVS( SkyboxVS_Input Input )
{
    SkyboxVS_Output Output;
    Output.Pos = mul( float4(Input.Pos, 1.0f), g_mWorldViewProj );
    Output.Tex = Input.Tex;
    return Output;
}

//-----------------------------------------------------------------------------
// Name: SkyboxPS
//-----------------------------------------------------------------------------
float4 SkyboxPS( SkyboxVS_Output Input ) : COLOR
{
    return tex2D( s0, Input.Tex );
}

//-----------------------------------------------------------------------------
// Name: SceneVS
//-----------------------------------------------------------------------------
struct SceneVS_Input
{
    float3 Pos : POSITION;
    float3 Normal : NORMAL;
};

struct SceneVS_Output
{
    float4 Pos : POSITION;					// Transformed position
    float3 Normal_World : TEXCOORD0;		// Normal (world space)
    float3 ViewVec_World : TEXCOORD1;		// View vector (world space)
};

SceneVS_Output SceneVS( SceneVS_Input Input )
{
    SceneVS_Output Output;
    Output.Pos = mul( float4(Input.Pos, 1.0f), g_mWorldViewProj );
    Output.Normal_World = mul( Input.Normal, (float3x3)g_mWorld );
    float4 pos_World = mul( float4(Input.Pos, 1.0f), g_mWorld );
    Output.ViewVec_World = pos_World.xyz - g_vEyePt;
    return Output;
}

//-----------------------------------------------------------------------------
// Name: ScenePS
//-----------------------------------------------------------------------------
float4 ScenePS( SceneVS_Output Input ) : COLOR
{
    // Sample the environment map
    float3 vReflect = reflect( Input.ViewVec_World, Input.Normal_World );
    float4 vEnvironment = texCUBE( CubeSampler, vReflect );
        
    // Simple overhead lighting
    float3 vColor = saturate( model_color * Input.Normal_World.y );
    
    // Add in reflection
    vColor = lerp( vColor, vEnvironment.rgb, MODEL_REFLECTIVITY );
  
	return float4( vColor, 1.0f );
}



//-----------------------------------------------------------------------------
// Name: FinalPass
//-----------------------------------------------------------------------------
float4 FinalPass( SkyboxVS_Output Input ) : COLOR
{
	float4 vColor = tex2D( s0, Input.Tex );
	float3 vBloom = tex2D( s1, Input.Tex );
	float vLum = g_fImageKey;
	vColor.rgb = vColor; 
	 
    // Tone mapping
    vColor.rgb *= g_fMiddleGray /(vLum + 0.001f);
    vColor.rgb *= (1.0f + vColor/LUM_WHITE);
	vColor.rgb /= (1.0f + vColor);
	
	vColor.rgb += 0.6f * vBloom;
	vColor.a = 1.0f;
	
	return vColor;
}

//-----------------------------------------------------------------------------
// Name: FinalPassEncoded
//-----------------------------------------------------------------------------
float4 FinalPassEncoded( SkyboxVS_Output Input ) : COLOR
{
    float4 vColor = tex2D( s0, Input.Tex );
    
    return vColor;
}


//-----------------------------------------------------------------------------
// Name: DownScale3x3
//-----------------------------------------------------------------------------
float4 DownScale3x3( in float2 vScreenPosition : TEXCOORD0 ) : COLOR
{
    float fAvg = 0.0f; 
    float4 vColor;
    
    for( int i = 0; i < 9; i++ )
    {
        // Compute the sum of color values
        vColor = tex2D( s0, vScreenPosition + g_avSampleOffsets[i] );
        fAvg = vColor;
    }
    
    // Divide the sum to complete the average
    fAvg /= 9;

    return vColor;
}

//-----------------------------------------------------------------------------
// Name: DownScale3x3_BrightPass
//-----------------------------------------------------------------------------
float4 DownScale3x3_BrightPass( in float2 vScreenPosition : TEXCOORD0 ) : COLOR
{
    float3 vColor = 0.0f;
    float  fLum;
    
    fLum = g_fImageKey;
    
    for( int i = 0; i < 9; i++ )
    {
        // Compute the sum of color values
        float4 vSample = tex2D( s0, vScreenPosition + g_avSampleOffsets[i] );
        vColor += vSample;
    }
    
    // Divide the sum to complete the average
    vColor /= 9;
 
    // Bright pass and tone mapping
    vColor = max( 0.0f, vColor - BRIGHT_THRESHOLD );
    vColor *= g_fMiddleGray / (fLum + 0.001f);  
    vColor *= (1.0f + vColor/LUM_WHITE);
	vColor /= (1.0f + vColor);
	
    return float4(vColor, 1.0f);
}


//-----------------------------------------------------------------------------
// Name: Bloom
//-----------------------------------------------------------------------------
float4 Bloom( in float2 vScreenPosition : TEXCOORD0 ) : COLOR
{
    float4 vSample = 0.0f;
    float4 vColor = 0.0f;
    float2 vSamplePosition;
    
    for( int iSample = 0; iSample < 15; iSample++ )
    {
        // Sample from adjacent points
        vSamplePosition = vScreenPosition + g_avSampleOffsets[iSample];
        vColor = tex2D(s0, vSamplePosition);
        
        vSample += g_avSampleWeights[iSample]*vColor;
    }
    
    return vSample;
}


//-----------------------------------------------------------------------------
// Technique: Bloom
//-----------------------------------------------------------------------------
technique Bloom
{
    pass p0
    {
        PixelShader = compile ps_2_0 Bloom();
    }
}




//-----------------------------------------------------------------------------
// Technique: Skybox
//-----------------------------------------------------------------------------
technique Skybox
{
    pass p0
    {
        VertexShader = compile vs_2_0 SkyboxVS();
        PixelShader = compile ps_2_0 SkyboxPS();
    }
}


//-----------------------------------------------------------------------------
// Technique: Scene
//-----------------------------------------------------------------------------
technique Scene
{
    pass p0
    {
        VertexShader = compile vs_2_0 SceneVS();
        PixelShader = compile ps_2_0 ScenePS();
    }
}




//-----------------------------------------------------------------------------
// Technique: FinalPass
//-----------------------------------------------------------------------------
technique FinalPass
{
    pass p0
    {
        PixelShader = compile ps_2_0 FinalPass();
    }
}




//-----------------------------------------------------------------------------
// Technique: FinalPass
//-----------------------------------------------------------------------------
technique FinalPassEncoded_RGB
{
    pass p0
    {
        PixelShader = compile ps_2_0 FinalPassEncoded();
    }
}




//-----------------------------------------------------------------------------
// Technique: DownScale3x3
//-----------------------------------------------------------------------------
technique DownScale3x3
{
    pass p0
    {
        PixelShader = compile ps_2_0 DownScale3x3();
    }
}




//-----------------------------------------------------------------------------
// Technique: DownScale3x3_BrightPass
//-----------------------------------------------------------------------------
technique DownScale3x3_BrightPass
{
    pass p0
    {
        PixelShader = compile ps_2_0 DownScale3x3_BrightPass();
    }
}

