//-----------------------------------------------------------------------------
// File: skybox.cpp
//
// Desc: Encapsulation of skybox geometry and textures
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------
#include "dxstdafx.h"
#include "skybox.h"

DWORD SKYBOX_VERTEX::FVF = D3DFVF_XYZ | D3DFVF_TEX1;


//-----------------------------------------------------------------------------
// Name: CSkybox
// Desc: Constructor
//-----------------------------------------------------------------------------
CSkybox::CSkybox()
{
    m_pTexPosX = NULL;
    m_pTexNegX = NULL;
    m_pTexPosY = NULL;
    m_pTexNegY = NULL;
    m_pTexPosZ = NULL;
    m_pTexNegZ = NULL;

    m_pVB = NULL;
    m_pd3dDevice = NULL;

    m_pTexCubeEnvMap = NULL;

    m_fSize = 1.0f;
}




//-----------------------------------------------------------------------------
// Name: Create
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::Create( LPDIRECT3DDEVICE9 pDevice, float fSize )
{
    Destroy();

    m_pd3dDevice = pDevice;
    m_pd3dDevice->AddRef();

    m_fSize = fSize;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Destroy
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::Destroy()
{
    if( m_pd3dDevice == NULL )
        return E_FAIL;

    SAFE_RELEASE( m_pd3dDevice );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InitDeviceObjects
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::InitDeviceObjects()
{
    if( m_pd3dDevice == NULL )
        return E_FAIL;

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::RestoreDeviceObjects()
{
    HRESULT hr = S_OK;

    if( m_pd3dDevice == NULL )
        return E_FAIL;

    hr = m_pd3dDevice->CreateVertexBuffer( 24 * sizeof(SKYBOX_VERTEX), 
                                           D3DUSAGE_WRITEONLY, SKYBOX_VERTEX::FVF, 
                                           D3DPOOL_DEFAULT, &m_pVB, NULL );
    if( FAILED(hr) )
        return hr;

    // Fill the vertex buffer
    SKYBOX_VERTEX* pVertex = NULL;
    hr = m_pVB->Lock( 0, 0, (void**)&pVertex, 0 );
    if( FAILED(hr) )
        return hr;

    // Set the texture coordinates
    for( int i=0; i < 6; i++ )
    {
        pVertex[i*4 + 0].tex = D3DXVECTOR2( 0, 0 );
        pVertex[i*4 + 1].tex = D3DXVECTOR2( 1, 0 );
        pVertex[i*4 + 2].tex = D3DXVECTOR2( 0, 1 );
        pVertex[i*4 + 3].tex = D3DXVECTOR2( 1, 1 );
    }

    // Positive X
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1, -1 );

    // Negative X
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1,  1 );

    // Positive Y
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1,  1 );
    
    // Negative Y
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1, -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1, -1 );

    // Positive Z
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1,  1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1,  1 );

    // Negative Z
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1, -1,  -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1, -1,  -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3( -1,  1,  -1 );
    (pVertex++)->pos = m_fSize * D3DXVECTOR3(  1,  1,  -1 );

    m_pVB->Unlock();


    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: Render
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::Render()
{
    if( m_pd3dDevice == NULL )
        return E_FAIL;

    m_pd3dDevice->SetStreamSource( 0, m_pVB, 0, sizeof(SKYBOX_VERTEX) );
    m_pd3dDevice->SetFVF( SKYBOX_VERTEX::FVF );
    m_pd3dDevice->SetSamplerState( 0, D3DSAMP_ADDRESSU, D3DTADDRESS_CLAMP );
    m_pd3dDevice->SetSamplerState( 0, D3DSAMP_ADDRESSV, D3DTADDRESS_CLAMP );
    m_pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER, D3DTEXF_POINT );
    
    m_pd3dDevice->SetTexture( 0, m_pTexPosX );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 0, 2 );

    m_pd3dDevice->SetTexture( 0, m_pTexNegX );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 4, 2 );

    m_pd3dDevice->SetTexture( 0, m_pTexPosY );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 8, 2 );

    m_pd3dDevice->SetTexture( 0, m_pTexNegY );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 12, 2 );

    m_pd3dDevice->SetTexture( 0, m_pTexPosZ );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 16, 2 );

    m_pd3dDevice->SetTexture( 0, m_pTexNegZ );
    m_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLESTRIP, 20, 2 );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::InvalidateDeviceObjects()
{
    SAFE_RELEASE( m_pVB );
    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::DeleteDeviceObjects()
{
    SAFE_RELEASE( m_pTexPosX );
    SAFE_RELEASE( m_pTexNegX );
    SAFE_RELEASE( m_pTexPosY );
    SAFE_RELEASE( m_pTexNegY );
    SAFE_RELEASE( m_pTexPosZ );
    SAFE_RELEASE( m_pTexNegZ );

    SAFE_RELEASE( m_pTexCubeEnvMap );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: CreateTexturesFromCross
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::CreateTexturesFromCross( LPDIRECT3DTEXTURE9 pTexCross )
{
    HRESULT hr = S_OK;

    if( m_pd3dDevice == NULL )
        return E_FAIL;

    if( pTexCross == NULL )
        return E_INVALIDARG;

    // Release old textures
    SAFE_RELEASE( m_pTexPosX );
    SAFE_RELEASE( m_pTexNegX );
    SAFE_RELEASE( m_pTexPosY );
    SAFE_RELEASE( m_pTexNegY );
    SAFE_RELEASE( m_pTexPosZ );
    SAFE_RELEASE( m_pTexNegZ );

    SAFE_RELEASE( m_pTexCubeEnvMap );

    D3DSURFACE_DESC desc;
    hr = pTexCross->GetLevelDesc( 0, &desc );
    if( FAILED(hr) )
        return hr;

    int size = desc.Width / 3;
    
    RECT rc = {0};
    
    // Positive X
    rc.left   = 2 * size;
    rc.top    = 1 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexPosX, rc );
    if( FAILED(hr) )
        return hr;

    // Negative X
    rc.left   = 0 * size;
    rc.top    = 1 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexNegX, rc );
    if( FAILED(hr) )
        return hr;

    // Positive Y
    rc.left   = 1 * size;
    rc.top    = 0 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexPosY, rc );
    if( FAILED(hr) )
        return hr;

    // Negative Y
	rc.left   = 1 * size;
    rc.top    = 2 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexNegY, rc );
    if( FAILED(hr) )
        return hr;

    // Positive Z
    rc.left   = 1 * size;
    rc.top    = 1 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexPosZ, rc );
    if( FAILED(hr) )
        return hr;

    // Negative Z
	rc.left   = 1 * size;
    rc.top    = 3 * size;
    rc.right  = rc.left + size;
    rc.bottom = rc.top + size;

    hr = CreateTextureFromRect( pTexCross, &m_pTexNegZ, rc );
    if( FAILED(hr) )
        return hr;

    // Create the cubic environment map
    hr = m_pd3dDevice->CreateCubeTexture( size, 1, 0, desc.Format, D3DPOOL_MANAGED, &m_pTexCubeEnvMap, NULL );
    if( FAILED(hr) )
        return hr;

    FillEnvMap();

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: CreateTextureFromRect
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::CreateTextureFromRect( LPDIRECT3DTEXTURE9 pTexSrc, LPDIRECT3DTEXTURE9* ppTexDest, RECT& rc )
{
    HRESULT hr = S_OK;

    D3DSURFACE_DESC desc;
    hr = pTexSrc->GetLevelDesc( 0, &desc );
    if( FAILED(hr) )
        return hr;

    int width = rc.right - rc.left;
    int height = rc.bottom - rc.top;

    // Positive X
    hr = m_pd3dDevice->CreateTexture( width, height, 1, 0, desc.Format, D3DPOOL_MANAGED, ppTexDest, NULL ); 
    if( FAILED(hr) )
        return hr;

    LPDIRECT3DSURFACE9 pSurfSrc = NULL;
    hr = pTexSrc->GetSurfaceLevel( 0, &pSurfSrc );
    if( FAILED(hr) )
        return hr;

    LPDIRECT3DSURFACE9 pSurfDest = NULL;
    hr = (*ppTexDest)->GetSurfaceLevel( 0, &pSurfDest );
    if( FAILED(hr) )
        return hr;

    hr = D3DXLoadSurfaceFromSurface( pSurfDest, NULL, NULL, pSurfSrc, NULL, &rc, D3DX_FILTER_TRIANGLE, 0 );
    if( FAILED(hr) )
        return hr;


    SAFE_RELEASE( pSurfDest );
    SAFE_RELEASE( pSurfSrc );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: FillEnvMap
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::FillEnvMap()
{
    HRESULT hr = S_OK;

    LPDIRECT3DTEXTURE9 apTexSrc[] = { m_pTexPosX, m_pTexNegX, m_pTexPosY, m_pTexNegY, m_pTexPosZ, m_pTexNegZ, };
    
    for( int i=0; i < 6; i++ )
    {     
        LPDIRECT3DSURFACE9 pSurfSrc = NULL;
        hr = apTexSrc[i]->GetSurfaceLevel( 0, &pSurfSrc );
        if( FAILED(hr) )
            return hr;

        LPDIRECT3DSURFACE9 pSurfDest = NULL;
        hr = m_pTexCubeEnvMap->GetCubeMapSurface( (D3DCUBEMAP_FACES)i, 0, &pSurfDest );
        if( FAILED(hr) )
            return hr;

        hr = D3DXLoadSurfaceFromSurface( pSurfDest, NULL, NULL, pSurfSrc, NULL, NULL, D3DX_FILTER_NONE, 0 );
        if( FAILED(hr) )
            return hr;

        SAFE_RELEASE( pSurfSrc );
        SAFE_RELEASE( pSurfDest );
    }

    LPDIRECT3DSURFACE9 pSurface = NULL;
    hr = m_pTexCubeEnvMap->GetCubeMapSurface( D3DCUBEMAP_FACE_NEGATIVE_Z, 0, &pSurface );
    if( FAILED(hr) )
        return hr;

    hr = RotateSurface180( pSurface );
    if( FAILED(hr) )
        return hr;

    SAFE_RELEASE( pSurface );

    return S_OK;
}




//-----------------------------------------------------------------------------
// Name: RotateSurface180
// Desc: 
//-----------------------------------------------------------------------------
HRESULT CSkybox::RotateSurface180( LPDIRECT3DSURFACE9 pSurface )
{
    HRESULT hr = S_OK;

    D3DSURFACE_DESC desc;
    hr = pSurface->GetDesc( &desc );
    if( FAILED(hr) )
        return hr;


    LPDIRECT3DSURFACE9 pSurfTemp = NULL;
    m_pd3dDevice->CreateOffscreenPlainSurface( desc.Width, desc.Height, 
                                               desc.Format, D3DPOOL_SCRATCH, 
                                               &pSurfTemp, NULL );
    if( FAILED(hr) )
        return hr;

    hr = D3DXLoadSurfaceFromSurface( pSurfTemp, NULL, NULL, pSurface, NULL, NULL, D3DX_FILTER_NONE, 0 );
    if( FAILED(hr) )
        return hr;

    D3DLOCKED_RECT rectSrc = {0};
    hr = pSurfTemp->LockRect( &rectSrc, NULL, D3DLOCK_READONLY );
    if( FAILED(hr) )
        return hr;

    D3DLOCKED_RECT rectDest = {0};
    hr = pSurface->LockRect( &rectDest, NULL, 0 );
    if (FAILED(hr) )
        return hr;

    DWORD dwBytesPerTexel = rectSrc.Pitch / desc.Width;
    BYTE* pSrc = (BYTE*) rectSrc.pBits;
    
    // Increment src to bottom right
    pSrc += (desc.Height * rectSrc.Pitch) - dwBytesPerTexel;

    for( UINT y=0; y < desc.Height; y++ )
    {
        BYTE* pDest = (BYTE*) rectDest.pBits;
        pDest += y * rectDest.Pitch;

        for( UINT x=0; x < desc.Width; x++ )
        {
            memcpy( pDest, pSrc, dwBytesPerTexel );
            pSrc -= dwBytesPerTexel;
            pDest += dwBytesPerTexel;
        }
    }
    pSurface->UnlockRect();
    pSurfTemp->UnlockRect();

    SAFE_RELEASE( pSurfTemp );

    return S_OK;
}
